/*
 
 Copyright (C) 2012 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of rRNAdust.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */


#include "interface.h"
#include "io.h"
#include "rRNAfilter.h"


struct parameters* interface(struct parameters* param,int argc, char *argv[])
{
	int c;
	int help = 0;
	
	if (argc < 2){
		usage();
		exit(0);
	}
		
	param = malloc(sizeof(struct parameters));
	assert(param!= 0);
	param->infiles = 0;
	param->infile = 0;
	param->outfile = 0;

	param->quiet_flag = 0;
	param->num_query = 800000;
	param->kmer_size = 2;
	param->errors = 2;
	param->afile = 0;
	param->server = 0;
	param->num_threads = 8;
	
	while (1){	 
		static struct option long_options[] ={
			{"server",0,0, 's'},
			{"threads",required_argument,0, 't'},
			{"quiet",0,0,'q'},
			{"help",0,0,'h'},
			{0, 0, 0, 0}
		};
		
		int option_index = 0;
		c = getopt_long_only (argc, argv,"qhe:o:a:t:",long_options, &option_index);
		
		if (c == -1){
			break;
		}
		
		switch(c) {
			case 0:
				
				break;
			case 's':
				param->server = 1;
				break;
			case 't':
				param->num_threads = atoi(optarg);
				break;
			case 'e':
				param->errors = atoi(optarg);
				break;
			case 'o':
				param->outfile = optarg;
				break;
			case 'a':
				param->afile = optarg;
				break;
			case 'q':
				param->quiet_flag = 1;
				break;
			case 'h':
				help = 1;
				break;
			case '?':
				exit(1);
				break;
			default:
				fprintf(stderr,"default\n\n\n\n");
				abort ();
		}
	}
	//fprintf(stderr,"Viterbi: %d\n",param->viterbi);
	
	if(help){
		usage();
		free(param);
		exit(0);
	}
	param->infile = malloc(sizeof(char*)* (argc-optind));
	assert(param->infile!=0);
	
	c = 0;
	while (optind < argc){
		param->infile[c] =  argv[optind++];
		c++;
	}
	param->infiles = c;
	return param;
}


void usage()
{
	fprintf(stdout, "\rRNAdust %0.2f, Copyright (C) 2011, 2012 Timo Lassmann <timolassmann@me.com>\n", VERSION);
	fprintf(stdout, "\n");
	fprintf(stdout, "Usage:   rRNAdust  <rRNA.fa> <input> > out.fa\n\n");
	fprintf(stdout, "Options: -o             STR    write non-rRNA sequences to file [STDOUT]\n");
	fprintf(stdout, "         -e             INT    number of errors allowed [2]\n");
	fprintf(stdout, "         -t             INT    number of threads [8]\n");
	fprintf(stdout, "         -a             STR    write rRNA sequences to file [NA]\n");
	fprintf(stdout, "         -s             NA     do not attempt to read from stdin  [NA]\n");
	fprintf(stdout, "\n");
	
	fprintf(stdout, "Removes all sequences that match with 2 or less errors to rRNA sequences.\nWorks on fasta, fastq and SAM/BAM input files if samtools is installed.\n\n");
	fprintf(stdout, "NOTE: if the input is in SAM / BAM format the output will be in SAM. Reads\ncorresponding to rRNA will be flagged as not passing QC (0x100).\n\n");
	fprintf(stdout, "Examples:\n\n");
	fprintf(stdout, "bzcat  ~/tmp/GNP_random_set_100000.fa.bz2   | ./rRNAdust ~/tmp/U13369.1.fa -e 3  -a rRNA.fa > out.fa \n\n");
	fprintf(stdout, "rRNAdust ~/tmp/U13369.1.fa  ~/tmp/GNP_random_set_100000.fa -e 2 > out.fa \n");
	fprintf(stdout, "\n");
}


void free_param(struct parameters* param)
{
	free(param->infile);
	free(param);
}

