/*
 
 Copyright (C) 2012 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of rRNAdust.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include "interface.h"
#include "nuc_code.h"
#include "io.h"
#include "misc.h"
#include "rRNAfilter.h" 


int main (int argc,char * argv[]) {
	struct parameters* param = 0;
	

	
	
	int i;
	int sam =0;
	init_nuc_code();
	
	param = interface(param,argc,argv);
	
	struct fasta* reference = 0;
	
	reference = get_fasta(reference,param->infile[0],0,param->server);
	
	if(param->infiles == 1){
		compare(reference,param,&read_fasta_fastq,0,0);
	}
	
	for(i = 1; i < param->infiles;i++){
		sam = 0;
		if(byg_end(".bam", param->infile[i])   == strlen(param->infile[i])){
			sam = 2;
		}else if(byg_end(".sam", param->infile[i])   == strlen(param->infile[i])){
			sam = 1;
		}else if (byg_end(".fa", param->infile[i])  == strlen(param->infile[i])){
			sam = 0;
		}else if (byg_end(".fq", param->infile[i])  == strlen(param->infile[i])){
			sam = 0;
		}else if (byg_end(".fastq", param->infile[i])  == strlen(param->infile[i])){
			sam = 0;
		}else if (byg_end(".fastaq", param->infile[i])  == strlen(param->infile[i])){
			sam = 0;

		}else if (byg_end(".fasta", param->infile[i])  == strlen(param->infile[i])){
			sam = 0;
		}else{
			sam = -1;
		}
		
		if(sam != -1){
			if(sam == 0){
				compare(reference,param,&read_fasta_fastq,i,sam);
			}else{
				compare(reference,param,&read_sam_chunk,i,sam);
			}
		}		
	}
	free_aln(reference);
	free_param(param);
}


void compare(struct fasta* reference,struct parameters* param,int (*fp)(struct read_info** ,struct parameters*,FILE* ),int file_num,int sam)
{
	
	struct read_info** ri = 0;
	char tmp_text[1000];
	
	char command[1000];
	char line[MAX_LINE];
	unsigned char* seq[4];
	
	int _MM_ALIGN16 lengths[4];
	int _MM_ALIGN16 errors[4];
	
	int i,j,c;
	int test = 1;
	int numseq;
	int excluded = 0;
	int done = 0;
	int reverse = 0;
	FILE* file = 0;
	FILE* afile = 0;
	FILE* foutfile = 0;
	ri = malloc(sizeof(struct read_info*) * param->num_query);
	assert(ri!=0);
	for(i = 0; i < param->num_query;i++){
		ri[i] = malloc(sizeof(struct read_info));
		assert(ri[i] != 0);
		ri[i]->seq = 0;
		ri[i]->name = 0;
		ri[i]->qual = 0;
		ri[i]->len = 0;
		ri[i]->cigar = 0;
		ri[i]->md = 0;
		ri[i]->priors = 0;// malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		
		ri[i]->RNAME = 0;
		ri[i]->POS = 0;
		ri[i]->MRNM = 0;
		ri[i]->MPOS = 0;
		ri[i]->ISIZE = 0;
		ri[i]->remain = 0;
		ri[i]->flag = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->hits = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->identity = malloc(sizeof(float)* (LIST_STORE_SIZE+1));
		assert(ri[i]->flag != 0);
		assert(ri[i]->identity != 0);
		assert(ri[i]->hits != 0);
		
	}
	
	
	if(param->outfile){
		if (!(foutfile = fopen(param->outfile, "w" ))){
			fprintf(stderr,"Cannot open file '%s'\n",param->outfile);
			exit(-1);
		}
	}else{
		foutfile = stdout;
	}
	
	if (!isatty(0) && param->server == 0){
		file = stdin;
	}else{
		if(sam == 2){
			command[0] = 0;
			strcat ( command, "samtools view -H ");     // Copy name into full name
			i = sprintf (tmp_text, "%s ", param->infile[file_num]);
			strcat ( command, tmp_text);
			//fprintf(stderr,"%s\n",command);
			if (!(file = popen(command, "r"))) {
				fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[file_num],command);
				exit(-1);
			}
			while(fgets(line, MAX_LINE, file)){
				fprintf(foutfile,"%s",line);
			}
			pclose(file);
			command[0] = 0;
			strcat ( command, "samtools view  ");     // Copy name into full name
			i = sprintf (tmp_text, "%s ", param->infile[file_num]);
			strcat ( command, tmp_text);
			//fprintf(stderr,"%s\n",command);
			if (!(file = popen(command, "r"))) {
				fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[file_num],command);
				exit(-1);
			}
		}else if(sam == 1){
			command[0] = 0;
			strcat ( command, "samtools view -SH ");     // Copy name into full name
			i = sprintf (tmp_text, "%s ", param->infile[file_num]);
			strcat ( command, tmp_text);
			fprintf(stderr,"%s\n",command);
			if (!(file = popen(command, "r"))) {
				fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[file_num],command);
				exit(-1);
			}
			while(fgets(line, MAX_LINE, file)){
				fprintf(foutfile,"%s",line);
			}
			pclose(file);
			command[0] = 0;
			strcat ( command, "samtools view  -S ");     // Copy name into full name
			i = sprintf (tmp_text, "%s ", param->infile[file_num]);
			strcat ( command, tmp_text);
			fprintf(stderr,"%s\n",command);
			if (!(file = popen(command, "r"))) {
				fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[file_num],command);
				exit(-1);
			}
		}else{
			if (!(file = fopen(param->infile[file_num] , "r" ))){
				fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[file_num]);
				exit(-1);
			}
		}
	}
	
	
	
	
	if(param->afile){
		if (!(afile = fopen(param->afile, "w" ))){
			fprintf(stderr,"Cannot open file '%s'\n",param->afile );
			exit(-1);
		}
	}
	
	//fprintf(stderr,"Threads: %d\n",param->num_threads);
	while ((numseq = fp(ri, param,file)) != 0){
		for(i = 0; i < numseq;i++){
			assert(  ri[i]->len  == strlen((char*)ri[i]->qual) );
		}
		
//#ifdef DEBUG
		fprintf(stderr,"%d read\n",numseq );
//#endif
		
		
		
		if(param->num_threads){
			start_threads(reference,param, ri, numseq);
			if(sam){
				for(i = 0; i < numseq;i++){
					if(ri[i]->a){
						print_sam(ri[i],foutfile,0);
						excluded++;
					}else{
						print_sam(ri[i],foutfile,1);
					}
				}
			}else{
				for(i = 0; i < numseq;i++){
					if(ri[i]->a){
						if(param->afile){
							print_seq(ri[i],afile);
						}
						excluded++;
					}else{
						print_seq(ri[i],foutfile);
					}
				}
			}
		}else{
			
			for(i = 0; i < numseq-4;i+=4){
				test = 1;
				reverse = 0;
				for(c = 0;c < 4;c++){
					errors[c] = 100000;
					
				}
				for(j =0; j < reference->numseq;j++){
					seq[0] = ri[i]->seq;
					seq[1] = ri[i+1]->seq;
					seq[2] = ri[i+2]->seq;
					seq[3] = ri[i+3]->seq;
					lengths[0] =  ri[i]->len;
					lengths[1] =  ri[i+1]->len;
					lengths[2] =  ri[i+2]->len;
					lengths[3] =  ri[i+3]->len;
					validate_bpm_sse(seq,lengths,reference->string +  reference->s_index[j],reference->s_index[j+1] - reference->s_index[j],4);
					for(c = 0;c < 4;c++){
						if(lengths[c] < errors[c]){
							errors[c] = lengths[c];
						}
					}
					
					
					seq[0] = reverse_complement2( ri[i]->seq,ri[i]->len);
					seq[1] = reverse_complement2( ri[i+1]->seq,ri[i+1]->len);
					seq[2] = reverse_complement2( ri[i+2]->seq,ri[i+2]->len);
					seq[3] = reverse_complement2( ri[i+3]->seq,ri[i+3]->len);
					lengths[0] =  ri[i]->len;
					lengths[1] =  ri[i+1]->len;
					lengths[2] =  ri[i+2]->len;
					lengths[3] =  ri[i+3]->len;
					validate_bpm_sse(seq,lengths,reference->string +  reference->s_index[j],reference->s_index[j+1] - reference->s_index[j],4);
					for(c = 0;c < 4;c++){
						if(lengths[c] < errors[c]){
							errors[c] = lengths[c];
						}
					}
					
					seq[0] = reverse_complement2( ri[i]->seq,ri[i]->len);
					seq[1] = reverse_complement2( ri[i+1]->seq,ri[i+1]->len);
					seq[2] = reverse_complement2( ri[i+2]->seq,ri[i+2]->len);
					seq[3] = reverse_complement2( ri[i+3]->seq,ri[i+3]->len);
				}
				
				
				for(c = 0;c < 4;c++){
					
					if(errors[c] <= param->errors){
						excluded++;
						if(sam){
							print_sam(ri[i+c],foutfile, 0);
						}else{
							if(param->afile){
								print_seq(ri[i+c],afile);
							}
						}
					}else{
						//ri[i+c]->seq = reverse_complement2( ri[i+c]->seq,ri[i+c]->len);
						if(sam){
							print_sam(ri[i+c],foutfile, 1);
						}else{
							print_seq(ri[i+c],foutfile);	
						}
					}
				}
				
				
				done+=4;
				if(done % 100000 == 0){
					fprintf(stderr,"%d done\n",done);
				}
			}
			while(i < numseq){
				test = 1;
				reverse = 0;
				for(j =0; j < reference->numseq;j++){
					//fprintf(stderr,"%s	%d\n",reference->sn[j],reference->s_index[j+1] - reference->s_index[j]);
					c = bpm(reference->string +  reference->s_index[j], ri[i]->seq,reference->s_index[j+1] - reference->s_index[j] , ri[i]->len, param->errors);
					if(c <= param->errors){
						test = 0;
						break;
					}
					
					
					
					//if(test){
					ri[i]->seq = reverse_complement2( ri[i]->seq,   ri[i]->len);
					c = bpm(reference->string +  reference->s_index[j], ri[i]->seq,reference->s_index[j+1] - reference->s_index[j] , ri[i]->len,param->errors);
					reverse = 1;
					if(c <= param->errors){
						ri[i]->seq = reverse_complement2( ri[i]->seq,   ri[i]->len);
						test = 0;
						break;
					}
					ri[i]->seq = reverse_complement2( ri[i]->seq,   ri[i]->len);
					//}
					
					
				}
				//test = 1;
				if(sam){
					print_sam(ri[i],foutfile,test);
				}else{
					if(test){
						//if(reverse){
						//	reverse_complement2( ri[i]->seq,   ri[i]->len);
						//}
						
						print_seq(ri[i],foutfile);
					}else{
						if(param->afile){
							print_seq(ri[i],afile);
						}
						excluded++;
						//fprintf(stderr,"EXCLUDED.... %d\n",c);
					}
				}
				i++;
			}
		}
		
	}
	
	fprintf(stderr,"Excluded: %d\n",excluded);
	if(param->outfile){
		fclose(foutfile);
	}
	
	for(i = 0; i < param->num_query;i++){
		free(ri[i]->flag); 
		free(ri[i]->hits);
		free(ri[i]->identity);
		
		if(ri[i]->seq){
			free(ri[i]->seq);
		}
		if(ri[i]->name){
			free(ri[i]->name);
		}
		if(ri[i]->qual){
			free(ri[i]->qual);
		}
		if(ri[i]->priors){
			free(ri[i]->priors);
		}
		
		if(ri[i]->cigar){
			free(ri[i]->cigar);
		}
		if(ri[i]->md){
			free(ri[i]->md);
		}
		
		if(ri[i]->RNAME){
			free(ri[i]->RNAME);
		}
		if(ri[i]->POS){
			free(ri[i]->POS);
		}
		if(ri[i]->MRNM){
			free(ri[i]->MRNM);
		}
		if(ri[i]->MPOS){
			free(ri[i]->MPOS);
		}
		if(ri[i]->ISIZE){
			free(ri[i]->ISIZE);
		}
		if(ri[i]->remain){
			free(ri[i]->remain);
		}
		
		free(ri[i]);
	}
	free(ri);
	if (!isatty(0)&& param->server == 0){
	
	}else{
		if(sam >= 1){
			pclose(file);
		}else{
			fclose(file);
		}
	}
	
	if(param->afile){
		fclose(afile);
	}
}


void start_threads(struct fasta* reference,struct parameters* param, struct read_info** ri, int numseq)
{
	struct thread_data* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data)* param->num_threads);
	assert(thread_data_array!=0);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	
	interval =  (int)((double)numseq /(double)param->num_threads);
	//interval = interval - (interval &0x3);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].reference = reference;
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
		
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].reference = reference;
	thread_data_array[param->num_threads - 1].start = (param->num_threads - 1)*interval;
	thread_data_array[param->num_threads - 1].end = numseq;
	thread_data_array[param->num_threads - 1].param = param;
	
	

	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, &compare_thread, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	free(thread_data_array);
}


void* compare_thread(void *threadarg)
{
	struct thread_data *data;
	data = (struct thread_data *) threadarg;
	
	struct read_info** ri = data->ri;
	struct fasta* reference = data->reference;
	const int start = data->start; 
	const int end = data->end;
	
	int error_cut = data->param->errors;
	
	int i,j,c;
	int test = 1;
	int reverse = 0;
	unsigned char* seq[4];
	
	int _MM_ALIGN16 lengths[4];
	int _MM_ALIGN16 errors[4];

	//fprintf(stderr,"%d	%d\n",start,end);
	for(i = start; i <= end-4;i+=4){
		test = 1;
		reverse = 0;
		for(c = 0;c < 4;c++){
			errors[c] = 100000;
			
		}
		for(j =0; j < reference->numseq;j++){
			seq[0] = ri[i]->seq;
			seq[1] = ri[i+1]->seq;
			seq[2] = ri[i+2]->seq;
			seq[3] = ri[i+3]->seq;
			lengths[0] =  ri[i]->len;
			lengths[1] =  ri[i+1]->len;
			lengths[2] =  ri[i+2]->len;
			lengths[3] =  ri[i+3]->len;
			validate_bpm_sse(seq,lengths,reference->string +  reference->s_index[j],reference->s_index[j+1] - reference->s_index[j],4);
			for(c = 0;c < 4;c++){
				if(lengths[c] < errors[c]){
					errors[c] = lengths[c];
				}
			}
			
			
			seq[0] = reverse_complement2( ri[i]->seq,ri[i]->len);
			seq[1] = reverse_complement2( ri[i+1]->seq,ri[i+1]->len);
			seq[2] = reverse_complement2( ri[i+2]->seq,ri[i+2]->len);
			seq[3] = reverse_complement2( ri[i+3]->seq,ri[i+3]->len);
			lengths[0] =  ri[i]->len;
			lengths[1] =  ri[i+1]->len;
			lengths[2] =  ri[i+2]->len;
			lengths[3] =  ri[i+3]->len;
			validate_bpm_sse(seq,lengths,reference->string +  reference->s_index[j],reference->s_index[j+1] - reference->s_index[j],4);
			for(c = 0;c < 4;c++){
				if(lengths[c] < errors[c]){
					errors[c] = lengths[c];
				}
			}
			
			seq[0] = reverse_complement2( ri[i]->seq,ri[i]->len);
			seq[1] = reverse_complement2( ri[i+1]->seq,ri[i+1]->len);
			seq[2] = reverse_complement2( ri[i+2]->seq,ri[i+2]->len);
			seq[3] = reverse_complement2( ri[i+3]->seq,ri[i+3]->len);
		}
		for(c = 0;c < 4;c++){
			if(errors[c] <= error_cut){
				ri[i+c]->a = 1;
			}
		}
	}

	while(i < end){
		//fprintf(stderr,"Looking at %d	%d	%d\n",i,start,end);
		test = 1;
		reverse = 0;
		for(j =0; j < reference->numseq;j++){
			//fprintf(stderr,"%s	%d\n",reference->sn[j],reference->s_index[j+1] - reference->s_index[j]);
			c = bpm(reference->string +  reference->s_index[j], ri[i]->seq,reference->s_index[j+1] - reference->s_index[j] , ri[i]->len, error_cut);
			if(c <= error_cut){
				test = 0;
				break;
			}
			//if(test){
			ri[i]->seq = reverse_complement2( ri[i]->seq,   ri[i]->len);
			c = bpm(reference->string +  reference->s_index[j], ri[i]->seq,reference->s_index[j+1] - reference->s_index[j] , ri[i]->len,error_cut);
			reverse = 1;
			if(c <= error_cut){
				ri[i]->seq = reverse_complement2( ri[i]->seq,   ri[i]->len);
				test = 0;
				break;
			}
			ri[i]->seq = reverse_complement2( ri[i]->seq,   ri[i]->len);
			//}
			
			
		}
		if(!test){
			ri[i]->a = 1;
		}
		i++;
	}

	pthread_exit((void *) 0);
}




